package filesystem

import java.nio.file.FileAlreadyExistsException
import java.nio.file.Path
import kotlin.io.path.createDirectory
import kotlin.io.path.createFile
import kotlin.io.path.writeText

class FSCreator {
    /**
     * Create entry, leaving existing folders' contents, but overwriting existing files.
     * @throws CyclicFolderException Cyclic folders cannot be created.
     * @throws DuplicateEntryNameException A folder or sub-folder contains entries with duplicate names.
     */
    @Throws(CyclicFolderException::class, DuplicateEntryNameException::class)
    fun create(
        entryToCreate: FSEntry,
        destination: String,
    ) {
        // No point in running anything if we know the input is invalid.
        if (entryToCreate is FSFolder) {
            if (entryToCreate.isCyclic()) {
                throw CyclicFolderException()
            }
            if (entryToCreate.deepHasDuplicateNames()) {
                throw DuplicateEntryNameException()
            }
        }

        val queue =
            ArrayDeque(
                listOf(
                    entryToCreate to Path.of(destination),
                ),
            )

        while (queue.isNotEmpty()) {
            val (entry, dest) = queue.removeFirst()
            val path = dest.resolve(entry.name)
            try {
                when (entry) {
                    is FSFile -> path.createFile()
                    is FSFolder -> path.createDirectory()
                }
            } catch (_: FileAlreadyExistsException) {
            } // Allow files/folders to already exist.
            when (entry) {
                is FSFile -> path.writeText(entry.content)
                is FSFolder -> queue.addAll(entry.entries.map { it to path })
            }
        }
    }
}

class CyclicFolderException : Exception("Cyclic FSFolders are not supported")

class DuplicateEntryNameException : Exception("Folder contains entries with duplicate names")